import 'dart:convert';

import 'package:flutter/foundation.dart';
import 'package:shared_preferences/shared_preferences.dart';

class Data extends ChangeNotifier {
  // Configuration available values
  List _availableParameters = ['level', 'size', 'skin'];

  List _availableLevelValues = ['easy', 'medium', 'hard', 'nightmare'];
  List _availableSizeValues = ['2x2', '3x2', '3x3', '4x4'];
  List _availableSkinValues = ['default', 'food', 'nature', 'monsters'];
  List _shufflableSkins = ['food', 'nature', 'monsters'];

  List get availableParameters => _availableParameters;
  List get availableLevelValues => _availableLevelValues;
  List get availableSizeValues => _availableSizeValues;
  List get availableSkinValues => _availableSkinValues;

  // Application default configuration
  String _parameterLevel = '';
  String _parameterLevelDefault = 'medium';
  String _parameterSize = '';
  String _parameterSizeDefault = '3x3';
  String _parameterSkin = '';
  String _parameterSkinDefault = 'default';

  // Application current configuration
  String get parameterLevel => _parameterLevel;
  String get parameterSize => _parameterSize;
  String get parameterSkin => _parameterSkin;

  // Game data
  bool _assetsPreloaded = false;
  bool _gameIsRunning = false;
  bool _animationInProgress = false;
  int _blockSizeVertical = 0;
  int _blockSizeHorizontal = 0;
  List _cells = [];
  List _cellsSolved = [];
  List _shuffledCellValues = [];
  int? _currentCellCol;
  int? _currentCellRow;
  int? _currentCellValue;
  bool _showConflicts = false;
  int _givenTipsCount = 0;
  String _currentState = '';

  void updateParameterLevel(String parameterLevel) {
    _parameterLevel = parameterLevel;
    notifyListeners();
  }

  int get blockSizeVertical => _blockSizeVertical;
  int get blockSizeHorizontal => _blockSizeHorizontal;
  void updateParameterSize(String parameterSize) {
    _parameterSize = parameterSize;
    _blockSizeHorizontal = int.parse(_parameterSize.split('x')[0]);
    _blockSizeVertical = int.parse(_parameterSize.split('x')[1]);
    notifyListeners();
  }

  void updateParameterSkin(String parameterSkin) {
    _parameterSkin = parameterSkin;
    notifyListeners();
  }

  String getParameterValue(String parameterCode) {
    switch (parameterCode) {
      case 'level':
        return _parameterLevel;
      case 'size':
        return _parameterSize;
      case 'skin':
        return _parameterSkin;
    }
    return '';
  }

  List getParameterAvailableValues(String parameterCode) {
    switch (parameterCode) {
      case 'level':
        return _availableLevelValues;
      case 'size':
        return _availableSizeValues;
      case 'skin':
        return _availableSkinValues;
    }
    return [];
  }

  void setParameterValue(String parameterCode, String parameterValue) async {
    switch (parameterCode) {
      case 'level':
        updateParameterLevel(parameterValue);
        break;
      case 'size':
        updateParameterSize(parameterValue);
        break;
      case 'skin':
        updateParameterSkin(parameterValue);
        break;
    }
    final prefs = await SharedPreferences.getInstance();
    prefs.setString(parameterCode, parameterValue);
  }

  void initParametersValues() async {
    final prefs = await SharedPreferences.getInstance();
    setParameterValue('level', prefs.getString('level') ?? _parameterLevelDefault);
    setParameterValue('size', prefs.getString('size') ?? _parameterSizeDefault);
    setParameterValue('skin', prefs.getString('skin') ?? _parameterSkinDefault);
  }

  String get currentState => _currentState;

  String computeCurrentGameState() {
    String cellsValues = '';
    String stringValues = '0123456789ABCDEFG';
    for (var rowIndex = 0; rowIndex < _cells.length; rowIndex++) {
      for (var colIndex = 0; colIndex < _cells[rowIndex].length; colIndex++) {
        cellsValues += stringValues[_cells[rowIndex][colIndex].value];
        cellsValues += _cells[rowIndex][colIndex].isFixed ? 'x' : ' ';
      }
    }

    var currentState = {
      'level': _parameterLevel,
      'size': _parameterSize,
      'skin': _parameterSkin,
      'tipsCount': _givenTipsCount,
      'showConflicts': _showConflicts,
      'boardValues': cellsValues,
      'shuffledCellValues': _shuffledCellValues,
    };

    return json.encode(currentState);
  }

  void saveCurrentGameState() async {
    if (_gameIsRunning) {
      _currentState = computeCurrentGameState();

      final prefs = await SharedPreferences.getInstance();
      prefs.setString('savedState', _currentState);
    } else {
      resetCurrentSavedState();
    }
  }

  void resetCurrentSavedState() async {
    _currentState = '';

    final prefs = await SharedPreferences.getInstance();
    prefs.setString('savedState', _currentState);
    notifyListeners();
  }

  void loadCurrentSavedState() async {
    final prefs = await SharedPreferences.getInstance();
    _currentState = prefs.getString('savedState') ?? '';
  }

  bool hasCurrentSavedState() {
    return (_currentState != '');
  }

  Map<String, dynamic> getCurrentSavedState() {
    if (_currentState != '') {
      Map<String, dynamic> savedState = json.decode(_currentState);
      if (savedState.isNotEmpty) {
        return savedState;
      }
    }
    return {};
  }

  bool get gameIsRunning => _gameIsRunning;
  void updateGameIsRunning(bool gameIsRunning) {
    _gameIsRunning = gameIsRunning;
    notifyListeners();
  }

  bool get assetsPreloaded => _assetsPreloaded;
  void updateAssetsPreloaded(bool assetsPreloaded) {
    _assetsPreloaded = assetsPreloaded;
  }

  List get cells => _cells;
  void updateCells(List cells) {
    _cells = cells;
    notifyListeners();
  }

  List get cellsSolved => _cellsSolved;
  void updateCellsSolved(List cells) {
    _cellsSolved = cells;
  }

  void shuffleCellValues() {
    int maxCellValue = 16;
    List values = new List<int>.generate(maxCellValue, (i) => i + 1);

    if (_shufflableSkins.contains(_parameterSkin)) {
      values.shuffle();
      print('Shuffled tiles values: ' + values.toString());
    }

    _shuffledCellValues = values;
  }

  void setShuffleCellValues(List values) {
    _shuffledCellValues = values;
  }

  int getTranslatedValueForDisplay(int originalValue) {
    return _shuffledCellValues[originalValue - 1];
  }

  int? get currentCellCol => _currentCellCol;
  set updateCurrentCellCol(int? currentCellCol) {
    _currentCellCol = currentCellCol;
    notifyListeners();
  }

  int? get currentCellRow => _currentCellRow;
  set updateCurrentCellRow(int? currentCellRow) {
    _currentCellRow = currentCellRow;
    notifyListeners();
  }

  int? get currentCellValue => _currentCellValue;
  set updateCurrentCellValue(int? currentCellValue) {
    _currentCellValue = currentCellValue;
    notifyListeners();
  }

  int get givenTipsCount => _givenTipsCount;
  increaseGivenTipsCount() {
    _givenTipsCount = _givenTipsCount + 1;
    saveCurrentGameState();
    notifyListeners();
  }

  setGivenTipsCount(int value) {
    _givenTipsCount = value;
    notifyListeners();
  }

  resetGivenTipsCount() {
    setGivenTipsCount(0);
  }

  selectCell(int? col, int? row) {
    _currentCellCol = col;
    _currentCellRow = row;
    _currentCellValue = null;
    if (row != null && col != null) {
      if (_cells[row][col].value != 0) {
        _currentCellValue = _cells[row][col].value;
      }
    }
    notifyListeners();
  }

  updateCellValue(int? col, int? row, int value) {
    if ((col != null) && (row != null)) {
      if (!_cells[row][col].isFixed) {
        _cells[row][col].value = value;

        saveCurrentGameState();
        notifyListeners();
      }
    }
  }

  bool get showConflicts => _showConflicts;
  void updateShowConflicts(bool showConflicts) {
    _showConflicts = showConflicts;
    notifyListeners();
  }

  void toggleShowConflicts() {
    updateShowConflicts(!showConflicts);
    saveCurrentGameState();
  }

  bool get animationInProgress => _animationInProgress;
  void updateAnimationInProgress(bool animationInProgress) {
    _animationInProgress = animationInProgress;
    notifyListeners();
  }

  void setAnimatedBackground(List animatedCellsPattern) {
    int boardSideLength = _blockSizeHorizontal * _blockSizeVertical;
    for (var row = 0; row < boardSideLength; row++) {
      for (var col = 0; col < boardSideLength; col++) {
        _cells[row][col].isAnimated = animatedCellsPattern[row][col];
      }
    }
    notifyListeners();
  }

  void resetAnimatedBackground() {
    int boardSideLength = _blockSizeHorizontal * _blockSizeVertical;
    for (var row = 0; row < boardSideLength; row++) {
      for (var col = 0; col < boardSideLength; col++) {
        _cells[row][col].isAnimated = false;
      }
    }
  }
}
